/*
 * $Id: hmcut.c,v 1.2 1995/07/30 19:20:47 traister Exp traister $
 * hmcut.c: function HeckbertMedianCut performs the cut on IMAGE
 *
 * Copyright (c) 1995, Joe Traister
 *
 * $Log: hmcut.c,v $
 * Revision 1.2  1995/07/30  19:20:47  traister
 * Updated copyright prior to first public release
 *
 * Revision 1.1  1995/05/13  01:38:16  traister
 * Initial revision
 *
 */
/*   This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include "vimage.h"

#define FS_SCALE 1024
#define MAX_COLORS 262144
#define HASH_SIZE 262144
#define hash(r, g, b) (((r)^((g)<<5)^((b)<<10))&0x3ffff)

typedef struct colorhist_tag {
  int count;
  UCHAR r, g, b;
} COLORHIST;

typedef struct colorhistlist_tag {
  int count;
  UCHAR r, g, b;
  struct colorhistlist_tag *next;
} COLORHISTLIST;

typedef struct box_tag {
  int index, colors, sum;
} BOX;

int redcompare(const void*, const void*);
int greencompare(const void*, const void*);
int bluecompare(const void*, const void*);
int sumcompare(const void*, const void*);
COLORHIST *histogram(IMAGE*, int, int*);
COLORHIST *histogram32K(IMAGE*, int*);
COLORHISTLIST *hash_image(IMAGE*, int, int*);
void freehash(COLORHISTLIST*);
int lookuphash(int r, int g, int b, COLORHISTLIST *hashtable);
void addtohash(int r, int g, int b, int value, COLORHISTLIST *hashtable);

int HeckbertMedianCut(IMAGE *image)
{
  UCHAR *newbits, *pP;
  COLORHIST *chv;
  COLORHISTLIST *hashtable;
  BOX *bv;
  int i, *thisrerr=NULL, *nextrerr=NULL, *thisgerr=NULL, *nextgerr=NULL;
  int *thisberr=NULL, *nextberr=NULL;
  int pixwidth, fs_direction=0, row, col, limitcol, sr, sg, sb, *temperr, err;
  int fcolors=0, boxes, colors, index, dist, newdist;
  int minr, maxr, ming, maxg, minb, maxb, lowersum, halfsum;
  int r, g, b, sum, bi;
  float bl, rl, gl;
  
  /* Build histogram */
  if (image->palsize == -15)
    chv = histogram32K(image, &fcolors);
  else
    chv = histogram(image, MAX_COLORS, &fcolors);

/*
  fprintf(stderr, "Histogram completed. %i colors\n", fcolors);
  for (i=0; i < fcolors; i++)
    fprintf(stderr, "Color %5i: %3i %3i %3i - %5i\n", i, chv[i].r,
	    chv[i].g, chv[i].b, chv[i].count);
*/
  
  bv = (BOX*)malloc(sizeof(BOX)*256);
  
  /* Set up first box */
  bv[0].index = 0;
  bv[0].colors = fcolors;
  bv[0].sum = image->width*image->height;
  boxes = 1;
  
  /* Main loop: split boxes until we have enough */
  while (boxes < 256) {
    
    /* find the first splittable box */
    for (bi=0; bi < boxes; bi++)
      if (bv[bi].colors >= 2)
	break;
    if (bi == boxes)
      break; /* no more boxes to split */
    index = bv[bi].index;
    colors = bv[bi].colors;
    sum = bv[bi].sum;
    
    /* find the boundaries of the box */
    minr = maxr = chv[index].r;
    ming = maxg = chv[index].g;
    minb = maxb = chv[index].b;
    for (i=1; i < colors; i++) {
      if (chv[index+i].r < minr) minr = chv[index+i].r;
      if (chv[index+i].r > maxr) maxr = chv[index+i].r;
      if (chv[index+i].g < ming) ming = chv[index+i].g;
      if (chv[index+i].g > maxg) maxg = chv[index+i].g;
      if (chv[index+i].b < minb) minb = chv[index+i].b;
      if (chv[index+i].b > maxb) maxb = chv[index+i].b;
    }
    
    /* sort by largest dimension (by luminosity) */
    rl = 0.299*(maxr-minr);
    gl = 0.587*(maxg-ming);
    bl = 0.114*(maxb-minb);
    if (rl >= gl && rl >= bl)
      qsort((void*)&(chv[index]), colors, sizeof(COLORHIST), redcompare);
    else if (gl >= bl)
      qsort((void*)&(chv[index]), colors, sizeof(COLORHIST), greencompare);
    else
      qsort((void*)&(chv[index]), colors, sizeof(COLORHIST), bluecompare);

    /* find the median based on pixel (not color) counts */
    lowersum = chv[index].count;
    halfsum = sum/2;
    for (i=1; i < colors-1; i++) {
      if (lowersum >= halfsum)
	break;
      lowersum += chv[index+i].count;
    }

    /* split the box and sort by pixel count */
    bv[bi].colors = i;
    bv[bi].sum = lowersum;
    bv[boxes].index = index+i;
    bv[boxes].colors = colors-i;
    bv[boxes].sum = sum-lowersum;
    boxes++;
    qsort((void*)bv, boxes, sizeof(BOX), sumcompare);
  }

  /* choose representative color from box (by average pixel) */
  for (bi = 0; bi < boxes; bi++) {
    r = g = b = sum = 0;
    index = bv[bi].index;
    colors = bv[bi].colors;
    for (i=0; i < colors; i++) {
      r += chv[index+i].r*chv[index+i].count;
      g += chv[index+i].g*chv[index+i].count;
      b += chv[index+i].b*chv[index+1].count;
      sum += chv[index+i].count;
    }
    r /= sum;
    if (r > 255) r = 255;
    g /= sum;
    if (g > 255) g = 255;
    b /= sum;
    if (b > 255) b = 255;
    image->palette[bi] = r|(g<<8)|(b<<16);
  }
  free(bv);
  free(chv);

  newbits = (UCHAR*)malloc(image->width*image->height);
  pixwidth = image->palsize == -15 ? 2 : 3;

  if (floyd) {
    /* Initialize Floyd-Steinberg error vectors */
    thisrerr = (int*)malloc((image->width+2)*sizeof(int));
    nextrerr = (int*)malloc((image->width+2)*sizeof(int));
    thisgerr = (int*)malloc((image->width+2)*sizeof(int));
    nextgerr = (int*)malloc((image->width+2)*sizeof(int));
    thisberr = (int*)malloc((image->width+2)*sizeof(int));
    nextberr = (int*)malloc((image->width+2)*sizeof(int));
    srandom((int)(time(0)^getpid()));
    for (i=0; i<image->width+2; i++) {
      thisrerr[i] = random()%(FS_SCALE*2)-FS_SCALE;
      thisgerr[i] = random()%(FS_SCALE*2)-FS_SCALE;
      thisberr[i] = random()%(FS_SCALE*2)-FS_SCALE;
      /* (random errors in [-1 .. 1]) */
    }
    fs_direction = 1;
  }

  hashtable = (COLORHISTLIST*)malloc(sizeof(COLORHISTLIST)*HASH_SIZE);
  for (i=0; i < HASH_SIZE; i++) {
    hashtable[i].count = -1;
    hashtable[i].next = NULL;
  }
  for (row=0; row < image->height; row++) {
    if (floyd) {
      memset(nextrerr, 0, sizeof(int)*(image->width+2));
      memset(nextgerr, 0, sizeof(int)*(image->width+2));
      memset(nextberr, 0, sizeof(int)*(image->width+2));
    }
    if (!floyd || fs_direction) {
      col = 0;
      limitcol = image->width;
      pP = image->bits+row*image->width*pixwidth;
    } else {
      col = image->width-1;
      limitcol = -1;
      pP = image->bits+row*image->width*pixwidth+(image->width-1)*pixwidth;
    }
    do {
      if (image->palsize == -24) {
	sr = *pP;
	sg = *(pP+1);
	sb = *(pP+2);
      } else {
	sr = ((*((short*)pP)&0x7C00)>>7)+7;
	sg = ((*((short*)pP)&0x03E0)>>2)+7;
	sb = ((*((short*)pP)&0x001F)<<3)+7;
      }
      if (floyd) {
	/* Apply Floyd-Steinberg error to pixel */
	sr += thisrerr[col+1]/FS_SCALE;
	sg += thisgerr[col+1]/FS_SCALE;
	sb += thisberr[col+1]/FS_SCALE;
	if (sr < 0) sr = 0;
	if (sr > 255) sr = 255;
	if (sg < 0) sg = 0;
	if (sg > 255) sg = 255;
	if (sb < 0) sb = 0;
	if (sb > 255) sb = 255;
      }

      /* Compute closest match */
      if ((index = lookuphash(sr, sg, sb, hashtable)) < 0) {
	dist = 2000000000;
	for (i=0; i < boxes; i++) {
	  newdist = (sr-(image->palette[i]&0xff))*
	    (sr-(image->palette[i]&0xff))+
	      (sg-((image->palette[i]&0xff00)>>8))*
		(sg-((image->palette[i]&0xff00)>>8))+
		  (sb-((image->palette[i]&0xff0000)>>16))*
		    (sb-((image->palette[i]&0xff0000)>>16));
	  if (newdist < dist) {
	    index = i;
	    dist = newdist;
	  }
	}
	addtohash(sr, sg, sb, index, hashtable);
      }
      *(newbits+row*image->width+col) = index;

      /* Propogate Floyd-Steinberg errors */
      if (floyd && fs_direction) {
	err = (sr-(image->palette[index]&0xff))*FS_SCALE;
	thisrerr[col+2] += (err*7)>>4;
	nextrerr[col] += (err*3)>>4;
	nextrerr[col+1] += (err*5)>>4;
	nextrerr[col+2] += err>>4;
	err = (sg-((image->palette[index]&0xff00)>>8))*FS_SCALE;
	thisgerr[col+2] += (err*7)>>4;
	nextgerr[col] += (err*3)>>4;
	nextgerr[col+1] += (err*5)>>4;
	nextgerr[col+2] += err>>4;
	err = (sb-((image->palette[index]&0xff0000)>>16))*FS_SCALE;
	thisberr[col+2] += (err*7)>>4;
	nextberr[col] += (err*3)>>4;
	nextberr[col+1] += (err*5)>>4;
	nextberr[col+2] += err>>4;
      } else if (floyd) {
	err = (sr-(image->palette[index]&0xff))*FS_SCALE;
	thisrerr[col] += (err*7)>>4;
	nextrerr[col+2] += (err*3)>>4;
	nextrerr[col+1] += (err*5)>>4;
	nextrerr[col] += err>>4;
	err = (sg-((image->palette[index]&0xff00)>>8))*FS_SCALE;
	thisgerr[col] += (err*7)>>4;
	nextgerr[col+2] += (err*3)>>4;
	nextgerr[col+1] += (err*5)>>4;
	nextgerr[col] += err>>4;
	err = (sb-((image->palette[index]&0xff0000)>>16))*FS_SCALE;
	thisberr[col] += (err*7)>>4;
	nextberr[col+2] += (err*3)>>4;
	nextberr[col+1] += (err*5)>>4;
	nextberr[col] += err>>4;
      }
      if (!floyd || fs_direction) {
	col++;
	pP += pixwidth;
      } else {
	col--;
	pP -= pixwidth;
      }
    } while (col != limitcol);
    if (floyd) {
      temperr = thisrerr;
      thisrerr = nextrerr;
      nextrerr = temperr;
      temperr = thisgerr;
      thisgerr = nextgerr;
      nextgerr = temperr;
      temperr = thisberr;
      thisberr = nextberr;
      nextberr = temperr;
      fs_direction = ! fs_direction;
    }
  }
  if (floyd) {
    free(thisrerr);
    free(nextrerr);
    free(thisgerr);
    free(nextgerr);
    free(thisberr);
    free(nextberr);
  }
  freehash(hashtable);

  image->palsize = boxes;
  free(image->bits);
  image->bits = newbits;
  return 0;
}

int redcompare(const void *ch1, const void *ch2)
{
  return (int)((COLORHIST*)ch1)->r-((COLORHIST*)ch2)->r;
}

int greencompare(const void *ch1, const void *ch2)
{
  return (int)((COLORHIST*)ch1)->g-((COLORHIST*)ch2)->g;
}

int bluecompare(const void *ch1, const void *ch2)
{
  return (int)((COLORHIST*)ch1)->b-((COLORHIST*)ch2)->b;
}

int sumcompare(const void *b1, const void *b2)
{
  return ((BOX*)b2)->sum-((BOX*)b1)->sum;
}

COLORHIST *histogram32K(IMAGE *image, int *fcolors)
{
  COLORHIST *chv;
  int *counts, i;
  UCHAR *pP;

  *fcolors = 0;
  counts = (int*)malloc(sizeof(int)*MAX_COLORS);
  memset(counts, 0, sizeof(int)*MAX_COLORS);
  for (pP=image->bits; pP-image->bits < image->width*image->height*2; pP+=2) {
    if (!counts[*((short*)pP)])
      *fcolors += 1;
    counts[*((short*)pP)]++;
  }
  chv = (COLORHIST*)malloc(sizeof(COLORHIST)*(*fcolors));
  *fcolors = 0;
  for (i=0; i < MAX_COLORS; i++)
    if (counts[i]) {
      chv[*fcolors].count = counts[i];
      chv[*fcolors].r = ((i&0x7c00)>>7)|7;
      chv[*fcolors].g = ((i&0x3e0)>>2)|7;
      chv[*fcolors].b = ((i&0x1f)<<3)|7;
      *fcolors += 1;
    }
  free(counts);
  return chv;
}

COLORHIST *histogram(IMAGE *image, int maxcolors, int *fcolors)
{
  COLORHISTLIST *cht, *chl;
  COLORHIST *chv;
  int i;
  
  cht = hash_image(image, maxcolors, fcolors);
  chv = (COLORHIST*)malloc(sizeof(COLORHIST)*(*fcolors));
  *fcolors = 0;
  for (i=0; i < HASH_SIZE; i++)
    for (chl=&cht[i]; chl; chl = chl->next)
      if (chl->count) {
	chv[*fcolors].count = chl->count;
	chv[*fcolors].r = chl->r;
	chv[*fcolors].g = chl->g;
	chv[*fcolors].b = chl->b;
	*fcolors += 1;
      }
  freehash(cht);
  return chv;
}

COLORHISTLIST *hash_image(IMAGE *image, int maxcolors, int *fcolors)
{
  COLORHISTLIST *hashtable, *chl;
  int i, clearbits = 0;
  UCHAR *pP;

  do {
    hashtable = (COLORHISTLIST*)malloc(sizeof(COLORHISTLIST)*HASH_SIZE);
    for (i=0; i < HASH_SIZE; i++) {
      hashtable[i].count = 0;
      hashtable[i].next = NULL;
    }
    *fcolors = 0;
    for (pP=image->bits;
	 pP-image->bits < image->width*image->height*3; 
	 pP += 3) {
      i = hash(*pP, *(pP+1), *(pP+2));
      if (hashtable[i].count) {
	for (chl = &hashtable[i]; chl; chl = chl->next)
	  if (chl->r == *pP && chl->g == *(pP+1) && chl->b == *(pP+2))
	    break;
	if (chl)
	  chl->count++;
	else {
	  chl = (COLORHISTLIST*)malloc(sizeof(COLORHISTLIST));
	  chl->count = 1;
	  chl->r = *pP;
	  chl->g = *(pP+1);
	  chl->b = *(pP+2);
	  chl->next = hashtable[i].next;
	  hashtable[i].next = chl;
	  *fcolors += 1;
	}
      } else {
	hashtable[i].count = 1;
	hashtable[i].r = *pP;
	hashtable[i].g = *(pP+1);
	hashtable[i].b = *(pP+2);
	*fcolors += 1;
      }
      if (*fcolors >= MAX_COLORS) {
	clearbits <<= 1;
	clearbits |= 1;
	for (i=0; i < image->width*image->height*3; i++)
	  image->bits[i] |= clearbits;
	freehash(hashtable);
	break;
      }
    }
  } while (*fcolors >= MAX_COLORS);
  return hashtable;
}

void freehash(COLORHISTLIST *cht)
{
  int i;
  COLORHISTLIST *chl, *chtmp;

  for (i=0; i < HASH_SIZE; i++) {
    chl=cht[i].next;
    while (chl) {
      chtmp = chl->next;
      free(chl);
      chl = chtmp;
    }
  }
  free(cht);
}

void addtohash(int r, int g, int b, int value, COLORHISTLIST *hashtable)
{
  int i=hash(r, g, b);
  COLORHISTLIST *chl=&hashtable[i];

  if (chl->count < 0) {
    chl->r = r;
    chl->g = g;
    chl->b = b;
    chl->count = value;
  } else {
    while (chl && (chl->r != r || chl->g != g || chl->b != b))
      chl = chl->next;
    if (!chl) {
      chl = (COLORHISTLIST*)malloc(sizeof(COLORHISTLIST));
      chl->r = r;
      chl->g = g;
      chl->b = b;
      chl->count = value;
      chl->next = hashtable[i].next;
      hashtable[i].next = chl;
    }
  }
}

int lookuphash(int r, int g, int b, COLORHISTLIST *hashtable)
{
  int i=hash(r, g, b);
  COLORHISTLIST *chl=&hashtable[i];

  if (hashtable[i].count < 0)
    return -1;
  while (chl && (chl->r != r || chl->g != g || chl->b != b))
    chl = chl->next;
  if (chl)
    return chl->count;
  else
    return -1;
}
