/*
 * $Id: vimage.c,v 1.4 1995/08/06 03:05:55 traister Exp traister $
 * vimage - an SVGA picture viewer for Linux using VGAlib 1.3
 * Copyright (c) 1995, Joseph J. Traister
 * 
 * $Log: vimage.c,v $
 * Revision 1.4  1995/08/06  03:05:55  traister
 * Added ESC exit to final wait loop in SlideShow()
 *
 * Revision 1.3  1995/07/30  19:20:47  traister
 * Updated copyright prior to first public release
 *
 * Revision 1.2  1995/05/13  22:40:17  traister
 * Added "break" in slide show when ESC hit.
 *
 * Revision 1.1  1995/05/13  01:40:52  traister
 * Initial revision
 *
 */

/*   This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <vga.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <values.h>
#include <getopt.h>
#include <limits.h>
#include <string.h>
#include "vimage.h"

#define VERSION "1.0"

#ifdef NCURSES_SUPPORT
#define OPTSTRING "m:l:d:qxra2fs:Rcv"
#else
#define OPTSTRING "m:l:d:qxra2fs:Rv"
#endif

char progname[PATH_MAX], modename[256]="";
int quiet=0, automode=0, pictime=6, twopass=0, floyd=0, slide=0, reverse=0;
int timer;
unsigned loopcount=1, printstats=0;

#ifdef NCURSES_SUPPORT
int filesel=0;
#endif

int randcompare(const void *, const void *);

int main(int argc, char **argv)
{
  char temp[PATH_MAX], **fnames=NULL, *chp;
  extern int optind;
  extern char *optarg;
  int fc=0, i, errval=0, randfiles=0, pipe_input=0, filetype;
  FILE *fpin;
  IMAGE image;

  SetID();
  strcpy(progname, argv[0]);
  optind = 1;
  while ((i=getopt(argc, argv, OPTSTRING)) != EOF) {
    switch (i) {
    case 'l':
      loopcount = atoi(optarg);
      break;
      
    case 'd':
      pictime = atoi(optarg);
      break;
      
    case 'q':
      quiet = 1;
      break;

    case 'x':
      printstats = 1;
      break;

    case 'r':
      reverse = 1;
      break;

    case 'a':
      automode = 1;
      break;
      
    case 'm':
      strcpy(modename, optarg);
      break;

    case '2':
      twopass = 1;
      break;

    case 'f':
      floyd = 1;
      break;

    case 'R':
      randfiles=1;
      break;

    case 's':
      if (!(fpin = fopen(optarg, "r")))
	Error(strerror(errno), optarg);
      do
	fgets(temp, PATH_MAX, fpin);
      while (temp[0] == '#');
      fc = atoi(temp);
      if (fc <= 0) {
	errno = EINVAL;
	Error("Bad file name count in slide file", temp);
      }
      fnames = (char**)malloc(sizeof(char*)*fc);
      if (!fnames)
	Error("Error allocating file name array", NULL);
      for (i=0; i < fc && !(feof(fpin)); i++) {
	do
	  fgets(temp, PATH_MAX, fpin);
	while (temp[0] == '#');
	fnames[i] = (char*)malloc(strlen(temp)+1);
	if (!fnames[i])
	  Error("Error allocating file name string", NULL);
	chp = memccpy(fnames[i], temp, '\n', PATH_MAX);
	if (chp)
	  *(chp-1) = '\0';
      }
      if (i < fc)
	fc = i;
      if (!fc) {
	errno = EINVAL;
	Error("No slide file names found", NULL);
      }
      fclose(fpin);
      slide = 1;
      break;

    case 'v':
      printf("%s: version %s\n", progname, VERSION);
      return 0;

#ifdef NCURSES_SUPPORT
    case 'c':
      filesel=1;
      break;
#endif
      
    case '?':
    default:
      errval++;
    }
  }

  if (optind == argc && !slide) {
    if (!isatty(STDIN_FILENO)) {
      pipe_input = 1;
      sprintf(temp, "/tmp/vimage%i", getpid());
      fnames = (char**)malloc(sizeof(char*));
      fnames[0] = (char*)malloc(strlen(temp)+1);
      strcpy(fnames[0], temp);
      fpin = fopen(temp, "w");
      if (!fpin)
	Error("Error opening temporary file", temp);
      while ((i=getchar()) != EOF)
	fputc(i, fpin);
      fc = 1;
      fclose(fpin);
    } else
      errval++;
  }

  if (errval) {
    fprintf(stderr,
	    "Usage: %s [options] [-s slide | file ...]\n"
	    "Options:\n"
            "  -R        Display pictures in random order\n"
            "  -a        Automatically choose video mode based on picture "
            "size\n"
	    "            and number of colors\n"
#ifdef NCURSES_SUPPORT
	    "  -c        Use full-screen file selector\n"
#endif
	    "  -f        Add Floyd-Steinberg dithering to Heckbert median "
            "cut\n"
	    "  -q        Quiet, do not display or break on errors\n"
	    "  -r        Show pictures in reverse order\n"
	    "  -x        No picture display, list image statistics instead\n"
	    "  -2        Use Heckbert median cut (2-pass) color palette"
	    " computation\n"
	    "  -d n      Delay {n} seconds between pictures (Default=6)\n"
	    "  -l n      Loop through pictures {n} times (Default=1)\n"
	    "  -m mode   Use video mode {mode}\n"
	    "  -s slide  Read filenames from file {slide}\n"
	    "  -v        Print version and quit\n"
	    "  Shows JPEG, Targa, PPM, PBM, PGM, GIF and SunRaster images\n",
	    progname);
    exit(-EINVAL);
  }

  if (!slide && !pipe_input) {
    fc = argc-optind;
    fnames = (char**)malloc(sizeof(char*)*fc);
    if (!fnames)
      Error(strerror(errno), NULL);
    for (i=0; i < fc; i++) {
      fnames[i] = (char*)malloc(strlen(argv[i+optind])+1);
      if (!fnames[i])
	Error(strerror(errno), NULL);
      strcpy(fnames[i], argv[i+optind]);
    }
  }

#ifdef NCURSES_SUPPORT
  if (filesel && fc > 1)
    return fileSelector(fnames, fc);
#endif

  if (printstats) {
    for (i=0; i < fc; i++) {
      if (!(fpin = fopen(fnames[i], "r"))) {
	errno=EINVAL;
        Error(strerror(errno), fnames[i]);
      }
      filetype = GetFileHead(fpin, &image, fnames[i]);
      if (!filetype) {
	errno=EINVAL;
	Error("Unknown image format", fnames[i]);
      }
      fclose(fpin);
      printf("%s: %ix%ix%i ", fnames[i], image.width, image.height, 
	     image.palsize > 0 ? image.palsize : 1<<(-image.palsize));
      switch (filetype) {
      case GIF:
	puts("GIF");
	break;
      case PNM:
	puts("PNM");
	break;
      case SUN:
	puts("Sun Rasterfile");
	break;
      case TARGA:
	puts("Targa");
	break;
#ifdef JPEG_SUPPORT
      case JPEG:
	puts("JPEG");
	break;
#endif
      }
    }
    return 0;
  }

  if (randfiles) {
    srand(time(0) ^ getpid());
    qsort(fnames, fc, sizeof(char*), randcompare);
  }
    
  if (fc == 1)
    pictime = 0;

  InitVGA(modename);
  Slideshow(fnames, fc);

  if (pipe_input)
    unlink(fnames[0]);
  return 0;
}

void Slideshow(char **fnames, int fc)
{
  FILE *fpin;
  int i, filetype, loop=0;
  IMAGE image;
  char c;

  timer=0;
  if (!reverse)
    i = 0;
  else
    i = fc-1;
  while (1) {
    image.bits = NULL;
    if (!(fpin = fopen(fnames[i], "r"))) {
      if (!quiet)
        Error(strerror(errno), fnames[i]);
      else
      	goto next;
    }
    filetype = GetFileHead(fpin, &image, fnames[i]);
    if (!filetype) {
      if (!quiet) {
	errno = EINVAL;
	Error("Unknown image format", fnames[i]);
      } else
	goto next;
    }
    switch (filetype) {
    case GIF:
      if (!ReadGIF(fpin, &image)) {
	if (!quiet) {
	  errno = EINVAL;
	  Error("Error reading GIF file", fnames[i]);
	} else
	  goto next;
      }
      break;
    case PNM:
      if (!ReadPNM(fpin, &image)) {
	if (!quiet) {
	  errno = EINVAL;
	  Error("Error reading PNM file", fnames[i]);
	} else
	  goto next;
      }
      break;
    case SUN:
      if (!ReadRasterfile(fpin, &image)) {
	if (!quiet) {
	  errno = EINVAL;
	  Error("Error reading Sun Rasterfile", fnames[i]);
	} else
	  goto next;
      }
      break;
    case TARGA:
      if (!ReadTarga(fpin, &image)) {
	if (!quiet) {
	  errno = EINVAL;
	  Error("Error reading Targa file", fnames[i]);
	} else
	  goto next;
      }
      break;
#ifdef JPEG_SUPPORT
    case JPEG:
      if (!ReadJPEG(fpin, &image)) {
	if (!quiet) {
	  errno = EINVAL;
	  Error("Error reading JPEG file", fnames[i]);
	} else
	  goto next;
      }
#endif
    }
    fclose(fpin);
    if (DisplayImage(&image, fnames[i])) {
      free(image.bits);
      VGAShutdown();
      return;
    }
    free(image.bits);

    if (fc == 1) {
      do {
	c = vga_getch();
      } while (c != '\n' && c != 'q');
      break;
    }

  next:
    if (!reverse) {
      i++;
      if (i >= fc) {
	loop++;
	if (loop >= loopcount)
	  break;
	i = 0;
      }
    } else {
      i--;
      if (i < 0) {
	loop++;
	if (loop >= loopcount)
	  break;
	i = fc-1;
      }
    }
  }
  if (fc > 1) {
    fc = time(0);
    while (time(0)-fc < pictime)
      if (vga_getkey() == 0x1b)
	break;
  }
  VGAShutdown();
}

/*
  Error: fatal error handler
*/
void Error(char *message, char *filename)
{
  VGAShutdown();
#ifdef NCURSES_SUPPORT
  CursesShutdown();
#endif
  if (filename)
    fprintf(stderr, "%s: %s: %s\n", progname, message, filename);
  else
    fprintf(stderr, "%s: %s\n", progname, message);
  exit(-errno);
}

int GetFileHead(FILE *fpin, IMAGE *image, char *filename)
{
  if (ReadGIFHead(fpin, image))
    return GIF;
  rewind(fpin);
  if (ReadRasterfileHead(fpin, image))
    return SUN;
  rewind(fpin);
  if (ReadPNMHead(fpin, image))
    return PNM;
  rewind(fpin);
  if (ReadTargaHead(fpin, image))
    return TARGA;
#ifdef JPEG_SUPPORT
  rewind(fpin);
  if (ReadJPEGHead(fpin, image))
    return JPEG;
#endif
  return 0;
}

int randcompare(const void *arg1, const void *arg2)
{
  if (rand() < RAND_MAX/2)
    return 1;
  else
    return -1;
}
